/* 
Console
version 1.0
by Kaldrum
Console command stuff and map/global hpu toggle
"Globalhpu on/off" in console to turn toggle global halfpipe utilities
"Lua.hpu = active" on level header to enable for a level.
handled in wallrun, but for organization: "Lua.multiwallrun = active" to enable multiple walljumps
*/

if HPU
	print("\x85 Didn't load repeat instance.")
	return
end

rawset(_G, "HPU", {})

CV_RegisterVar{
	name = "GlobalHPU", 
	defaultvalue = 1, 
	flags = CV_NETVAR,
	PossibleValue = CV_OnOff
}

HPU.maphpu = false

HPU.Global = CV_FindVar("GlobalHPU")

addHook("MapChange", function(map)
	if mapheaderinfo[map].hpu == "active"
		HPU.active = true
		HPU.maphpu = true
	else
		HPU.active = false
		HPU.maphpu = false
	end
end)

addHook("ThinkFrame", function()
	if HPU.Global.value and not HPU.active
		print("Global Halfpipe Utilities is enabled.")
		HPU.active = true
	elseif not HPU.Global.value and HPU.active and not HPU.maphpu
		print("Global Halfpipe Utilities is disabled.")
		HPU.active = false
	end
end)

/*
Ceiling Run
version 1.2.2
by Lach
modified by Kaldrum
A modification of Lach's ArcadePhysics ceiling running script from Chrispy's Arcade. Should allow for implementation of loops.
This script implements temporary ceiling running, based on maintenance of a high enough movement speed. Can perform most movement options, but will fall when leaving the ground or moving too slow.
Requires landing on a slope to start ceiling running, and depending on player orientation when entering a slope will flip movement when upside down.
A player with flipcam enabled will flip actual orientation when this occurs, while a player without will have inverted "forward" movement controls.
No flipcam is the intended experience, however flipcam should be accomodated properly, as far more people use flipcam than not.
*/

if CeilingRun
	return
end

rawset(_G, "CeilingRun", true)

local CEILING_SPEED_THRESHOLD = 16*FRACUNIT

local function FixedSquare(n)
	return FixedMul(n, n)
end


//Will swap the current gravity state, allowing for ceiling running in existing reverse gravity.
local function FlipGravity(mo)
	mo.flags2 = $ ^^ MF2_OBJECTFLIP
	mo.eflags = $ ^^ MFE_VERTICALFLIP
end

//Cool slope math (that I don't understand)
local function FV3_Rotate(rotVecX, rotVecY, rotVecZ, axisVecX, axisVecY, axisVecZ, angle)
	local ux, uy, uz = FixedMul(axisVecX, rotVecX), FixedMul(axisVecX, rotVecY), FixedMul(axisVecX, rotVecZ)
	local vx, vy, vz = FixedMul(axisVecY, rotVecX), FixedMul(axisVecY, rotVecY), FixedMul(axisVecY, rotVecZ)
	local wx, wy, wz = FixedMul(axisVecZ, rotVecX), FixedMul(axisVecZ, rotVecY), FixedMul(axisVecZ, rotVecZ)
	local sa, ca = sin(angle), cos(angle)
	local ua = ux + vy + wz
	local ax, ay, az = FixedMul(axisVecX, ua), FixedMul(axisVecY, ua), FixedMul(axisVecZ, ua)
	local xs, ys, zs = FixedSquare(axisVecX), FixedSquare(axisVecY), FixedSquare(axisVecZ)
	local bx, by, bz = FixedMul(rotVecX, ys + zs), FixedMul(rotVecY, xs + zs), FixedMul(rotVecZ, xs + ys)
	local cx, cy, cz = FixedMul(axisVecX, vy + wz), FixedMul(axisVecY, ux + wz), FixedMul(axisVecZ, ux + vy)
	local dx, dy, dz = FixedMul(bx - cx, ca), FixedMul(by - cy, ca), FixedMul(bz - cz, ca)
	local ex, ey, ez = FixedMul(vz - wy, sa), FixedMul(wx - uz, sa), FixedMul(uy - vx, sa)
	return ax + dx + ex, ay + dy + ey, az + dz + ez
end

local function GetSlopeLandingMomentum(mo, slope)
	if not slope
	or slope.flags & SL_NOPHYSICS
	or slope.zangle == 0
		return mo.momx, mo.momy
	end
	
	return FV3_Rotate(mo.momx, mo.momy, 2 * mo.momz, -slope.d.y, slope.d.x, 0, InvAngle(slope.zangle))
end

//Gets the slope of the floor above the player. 
local function GetFloorSlope(mo)
	if not (mo.eflags & MFE_VERTICALFLIP) == not true
		local fof = mo.floorrover
		if fof
			return fof.t_slope
		end
		return mo.subsector.sector.f_slope
	else
		local fof = mo.ceilingrover
		if fof
			return fof.b_slope
		end
		return mo.subsector.sector.c_slope
	end
end

local function GetFloorSlopeZ(mo)
	local slope = mo.standingslope
	if slope == nil or slope.flags & SL_NOPHYSICS
		return 0
	else
		return slope.zangle
	end
end

local function GetFloorSlopeXY(mo, slope)
	local zangle = slope.zangle
	local xyangle = slope.xydirection
	if (slope.flags & SL_NOPHYSICS) or (zangle == 0)
		return nil
	end
	if zangle < 0 
		xyangle = $ + ANGLE_180
	end
	if mo.eflags & MFE_VERTICALFLIP
		xyangle = $ + ANGLE_180
	end
	return xyangle
end

//Determines if the object will touch the floor based on their height and momentum
local function MobjWillTouchFloor(mo)
	if not (mo.eflags & MFE_VERTICALFLIP) == not true
		return mo.z + mo.momz < mo.floorz
	else
		return mo.z + mo.height + mo.momz > mo.ceilingz
	end
end

local function P_QuantizeMomentumToSlope(vx, vy, vz, slope, slopeangle)
	local ax = 0
	local ay = 0
	local az = 0
	if (slope.flags & SL_NOPHYSICS)
		return
	end
	
	ax = -slope.d.y
	ay = slope.d.x
	az = 0
	return FV3_Rotate(vx, vy, vz, ax, ay, az, slopeangle)
end

local function P_ShouldResetConveyorMomentum(player)
	if player.onconveyor == 1
		return false
	elseif player.onconveyor == 2
		return not(player.mo.eflags & (MFE_UNDERWATER|MFE_TOUCHWATER))
	elseif player.onconveyor == 3
		return true
	elseif player.onconveyor == 4
		return not P_IsObjectOnGround(player.mo)
	else
		return true
	end
end

//undoes the slope physics	
local function P_ReverseButteredSlope(mo, momx, momy)
	local thrust = 0
	if not mo.standingslope
		return
	end
	
	if mo.standingslope.flags & SL_NOPHYSICS
		return
	end
	
	if mo.flags & (MF_NOCLIPHEIGHT|MF_NOGRAVITY)
		return
	end
	
	if mo.player
		if (abs(mo.standingslope.zdelta) < FRACUNIT/4 and not(mo.player.pflags & PF_SPINNING))
			return
		end
		if (abs(mo.standingslope.zdelta) < FRACUNIT/2 and not(mo.player.rmomx or mo.player.rmomy))
			return
		end
	end
	thrust = sin(mo.standingslope.zangle) * 3/2 * -P_MobjFlip(mo)
	if mo.player and (mo.player.pflags & PF_SPINNING)
		local mult = 0
		if momx or momy
			local angle = R_PointToAngle2(0,0,momx, momy) - mo.standingslope.xydirection
			if (P_MobjFlip(mo) * mo.standingslope.zdelta < 0)
				angle = $+ANGLE_180
			end
			mult = cos(angle)
		end
		thrust = FixedMul(thrust, FRACUNIT*2/3 + mult/8)
	end
	if momx or momy
		thrust = FixedMul(thrust, FRACUNIT + P_AproxDistance(momx, momy)/16)
	end
	
	thrust = FixedMul(thrust, abs(P_GetMobjGravity(mo)))
	
	thrust = FixedMul(thrust, FixedDiv(mo.friction, 29*FRACUNIT/32))
	
	P_Thrust(mo, mo.standingslope.xydirection, -2 * thrust)
end

//slightly modified movement code
local function P_3dMovement(player, momx, momy)

	local movepushangle, movepushsideangle = 0, 0
	local topspeed, acceleration, thrustfactor = 0, 0, 0
	local movepushforward, movepushside = 0, 0
	local mforward, mbackward = 0, 0
	local dangle = 0
	local normalspeed = FixedMul(player.normalspeed, player.mo.scale)
	local spin = (P_IsObjectOnGround(player.mo) and (player.pflags & (PF_SPINNING|PF_THOKKED)) == PF_SPINNING and (player.rmomx or player.rmomy) and not(player.pflags & PF_STARTDASH))
	local oldmagnitude, newmagnitude = 0, 0
	local totalthrustx, totalthrusty, totalthrustz = 0, 0, 0
	local momx, momy = player.mo.momx, player.mo.momy
	local cmomx, cmomy = player.cmomx, player.cmomy
	local rmomx, rmomy = 0, 0
	local speed = 0
	local aiming = 0
	
	totalthrustz = FRACUNIT*P_MobjFlip(player.mo)/3
	
	oldmagnitude = R_PointToDist2(momx - cmomx, momy - cmomy, 0, 0)
	
	if player.pflags & PF_ANALOGMODE
		movepushangle = player.cmd.angleturn << 16
	else
		movepushangle = player.mo.angle
	end
	
	movepushsideangle = movepushangle - ANGLE_90
	
	if P_ShouldResetConveyorMomentum(player)
		cmomx = 0
		cmomy = 0
	end
	rmomx = momx - cmomx
	rmomy = momy - cmomy
	
	speed = P_AproxDistance(rmomx, rmomy)    
	
	dangle = R_PointToAngle2(0, 0, player.rmomx, player.rmomy) - player.mo.angle
	if dangle > ANGLE_180
		dangle = InvAngle(dangle)
	end
	
	if dangle <= ANGLE_45
		mforward = 1
	elseif dangle >= ANGLE_135
		mbackward = 1
	end
	
	if player.pflags & PF_SLIDING
		player.cmd.forwardmove = 0
	elseif P_IsObjectOnGround(player.mo) and player.mo.state == S_PLAY_PAIN
		player.mo.state = S_PLAY_WALK
	end
	
	aiming = player.cmd.aiming << 16
	
	if player.pflags & PF_SLIDING
		normalspeed = FixedMul(36<<16, player.mo.scale)
		thrustfactor = 5
		acceleration = 96 + (FixedDiv(speed, player.mo.scale)>>16) * 40
		topspeed = normalspeed
	else
		if player.powers[pw_super] or player.powers[pw_sneakers]
			topspeed = 5 * normalspeed/3
			thrustfactor = player.thrustfactor * 2
			acceleration = player.accelstart/2 + (FixedDiv(speed, player.mo.scale)>>16) * player.acceleration/2
		else
			topspeed = normalspeed
			thrustfactor = player.thrustfactor
			acceleration = player.accelstart + (FixedDiv(speed, player.mo.scale)>>16) * player.acceleration
		end
		
		if player.powers[pw_tailsfly]
			topspeed = $>>1
		elseif player.mo.eflags & (MFE_UNDERWATER|MFE_GOOWATER)
			topspeed = $>>1
			acceleration = 2*acceleration/3
		end
	end
	if spin
		local ns = FixedDiv(549*29*FRACUNIT/32, 500* FRACUNIT)
		topspeed = FixedMul(oldmagnitude, ns)
	end

	if not player.powers[pw_tailsfly]
		if player.pflags & PF_BOUNCING
			if player.mo.state == S_PLAY_BOUNCE_LANDING
				thrustfactor = player.thrustfactor * 8
				acceleration = player.accelstart/8 + (FixedDiv(speed, player.mo.scale)>>16) * player.acceleration/8
			else
				thrustfactor = (3*player.thrustfactor)/4
				acceleration = player.accelstart + (FixedDiv(speed, player. mo.scale)>>16) * player.acceleration
			end
		end
		
		if player.mo.movefactor ~= FRACUNIT
			acceleration = FixedMul(acceleration<<16, player.mo.movefactor)>>16
		end
	end
	
	if player.climbing
	
	elseif not (player.pflags & PF_ANALOGMODE) and player.cmd.forwardmove ~= 0 and not(player.pflags & PF_GLIDING or player.exiting or P_PlayerInPain(player) and not P_IsObjectOnGround(player.mo))
		
		movepushforward = player.cmd.forwardmove * (thrustfactor * acceleration)
		
		if (player.pflags & (PF_SPINNING|PF_THOKKED)) == PF_SPINNING
			if (mforward and player.cmd.forwardmove > 0) or (mbackward and player.cmd.forwardmove < 0) or (player.pflags & PF_STARTDASH)
				movepushforward = 0
			elseif P_IsObjectOnGround(player.mo)
				movepushforward = $ / 16
			else
				movepushforward = $ /8
			end
		elseif not P_IsObjectOnGround(player.mo)
			movepushforward = $ / 4
		end
		movepushforward = FixedMul(movepushforward, player.mo.scale)
		
		totalthrustx = $ + P_ReturnThrustX(player.mo, movepushangle, movepushforward)
		totalthrusty = $ + P_ReturnThrustY(player.mo, movepushangle, movepushforward)	
	end
	
	if player.climbing
	
	elseif (player.pflags & PF_ANALOGMODE)
		if not(player.pflags & PF_GLIDING or player.exiting or P_PlayerInPain(player))
			local controldirection = R_PointToAngle2(0, 0, player.cmd.forwardmove * FRACUNIT, -player.cmd.sidemove * FRACUNIT) + movepushangle
			movepushforward = FixedHypot(player.cmd.sidemove, player.cmd.forwardmove) * (thrustfactor * acceleration)
			
			if (player.pflags & (PF_SPINNING|PF_THOKKED)) == PF_SPINNING
				if (mforward and player.cmd.forwardmove > 0) or (mbackward and player.cmd.forwardmove < 0) or (player.pflags & PF_STARTDASH)
					movepushforward = 0
				elseif P_IsObjectOnGround(player.mo)
					movepushforward = $ >> 4
				else
					movepushforward = $ >> 3
				end
			elseif not P_IsObjectOnGround(player.mo)
				movepushforward = $ >> 2
			end
			
			movepushsideangle = controldirection
			
			movepushforward = FixedMul(movepushforward, player.mo.scale)
			
			totalthrustx = $ + P_ReturnThrustX(player.mo, controldirection, movepushforward)
			totalthrusty = $ + P_ReturnThrustY(player.mo, controldirection, movepushforward)	
		end
	elseif player.cmd.sidemove and not (player.pflags & PF_GLIDING) and not player.exiting and not P_PlayerInPain(player)
		movepushside = player.cmd.sidemove * (thrustfactor * acceleration)
		
		if not P_IsObjectOnGround(player.mo)
			movepushside = $ / 4
			
			if ((player.pflags & (PF_SPINNING|PF_THOKKED)) == PF_SPINNING) or (player.powers[pw_tailsfly] and speed > topspeed)
				movepushside = $ / 4
			end
		elseif (player.pflags & (PF_SPINNING|PF_THOKKED)) == PF_SPINNING
			if player.pflags & PF_STARTDASH
				movepushside = 0
			elseif P_IsObjectOnGround(player.mo)
				movepushside = $ / 16
			else
				movepushside = $ / 8
			end
		end
		
		movepushside = FixedMul(movepushside, player.mo.scale)
		
		totalthrustx = $ + P_ReturnThrustX(player.mo, movepushsideangle, movepushside)
		totalthrusty = $ + P_ReturnThrustY(player.mo, movepushsideangle, movepushside)
	end
	
	if (totalthrustx or totalthrusty) and player.mo.standingslope and (not(player.mo.standingslope.flags & SL_NOPHYSICS)) and abs(player.mo.standingslope.zdelta) > FRACUNIT/2
		local thrustangle = R_PointToAngle2(0, 0, totalthrustx, totalthrusty)-player.mo.standingslope.xydirection
		//this is different in order to reduce movement on downwards slopes instead of on upwards ones as it would normally do
		if player.mo.standingslope.zdelta > 0
			if (thrustangle < ANGLE_90) or (thrustangle > ANGLE_270)
				local momentum = {P_QuantizeMomentumToSlope(totalthrustx, totalthrusty, totalthrustz, player.mo.standingslope, player.mo.standingslope.zangle)}
				totalthrustx = momentum[1]
				totalthrusty = momentum[2]
			end
		else
			if (thrustangle > ANGLE_90) and (thrustangle < ANGLE_270)
				local momentum = {P_QuantizeMomentumToSlope(totalthrustx, totalthrusty, totalthrustz, player.mo.standingslope, player.mo.standingslope.zangle)}
				totalthrustx = momentum[1]
				totalthrusty = momentum[2]
			end
		end
	end
	
	if totalthrustx or totalthrusty 
		momx = $ + totalthrustx
		momy = $ + totalthrusty
	else
		return nil
	end
	
	newmagnitude = R_PointToDist2(momx - cmomx, momy - cmomy, 0, 0)
	if newmagnitude > topspeed
		local tempmomx, tempmomy = 0, 0
		if (oldmagnitude > topspeed) and (not spin)
			if newmagnitude > oldmagnitude
				tempmomx = FixedMul(FixedDiv(momx - cmomx, newmagnitude), oldmagnitude)
				tempmomy = FixedMul(FixedDiv(momy - cmomy, newmagnitude), oldmagnitude)
				momx = tempmomx + cmomx
				momy = tempmomy + cmomy
			end
		else
			tempmomx = FixedMul(FixedDiv(momx - cmomx, newmagnitude), topspeed)
			tempmomy = FixedMul(FixedDiv(momy - cmomy, newmagnitude), topspeed)
			momx = tempmomx + cmomx
			momy = tempmomy + cmomy
		end
	end
	return momx, momy, rmomx, rmomy, cmomx, cmomy, speed
end

//Sets variables relating to modified movement physics
local function Set3DMoveVars(player)
	local movevars = {P_3dMovement(player)}
	if movevars[1] ~= nil
		player.nmomx = movevars[1]
		player.nmomy = movevars[2]
		player.nrmomx = movevars[3]
		player.nrmomy = movevars[4]
		player.ncmomx = movevars[5]
		player.ncmomy = movevars[6]
		player.nspeed = movevars[7]
	else
		player.nmomx = nil
		player.nmomy = nil
		player.nrmomx = nil
		player.nrmomy = nil
		player.ncmomx = nil
		player.ncmomy = nil
		player.nspeed = nil
	end
end

addHook("PreThinkFrame", function()
	if HPU.active == false
		return
	end
	
	for player in players.iterate do 
		local mo = player.mo
		if not mo
			continue
		end
		
		player.lastslopeang = GetFloorSlopeZ(mo)
		
		if not WallRun and mo.standingslope
			local dir = GetFloorSlopeXY(mo, mo.standingslope)
			if dir ~= nil
				player.lastslope = dir
			end
		end
		
		if player.ceilingrun and not(twodlevel or mo.flags2 & MF2_TWOD)
			Set3DMoveVars(player)
		end
		
		//Cannot run on the ceiling if clipping through ceilings or floors, or carrying something.
		if (mo.flags & MF_NOCLIPHEIGHT)or (player.powers[pw_carry] ~= CR_NONE)
			if player.ceilingrun
				FlipGravity(mo)
				player.ceilingrun = false
				player.flipdir = false
				player.falltimer = 0
			end
			continue
		end
		if P_IsObjectOnGround(mo)
			if not player.ceilingrun
				continue
			end
			
			if player.wasroll and (mo.state ~= S_PLAY_ROLL) and not(player.cmd.buttons & BT_SPIN) and flip
				if player.cmd.forwardmove < 0
					mo.angle = player.lastslope
				else
					mo.angle = player.lastslope + ANGLE_180
				end
				player.cmd.forwardmove = -$
				player.wasroll = 0
			elseif player.wasroll and (mo.state == S_PLAY_ROLL)
				//depletes a timer for every tic spent rolling, should cover the entire slope transfer. 
				//After this runs out, any actions that remove spin somehow will not flip the camera angle. Largely should prevent issues with uncurls when upside down
				player.wasroll = $-1
			end
			
			//cancels ceilingrun if too slow
			if FixedHypot(mo.momx - player.cmomx , mo.momy - player.cmomy) < FixedMul(10*FRACUNIT, mo.scale)
				FlipGravity(mo)
				player.ceilingrun = false
				player.flipdir = false
				player.falltimer = 0
				player.wasroll = 0
			end
			continue
		end
		
		//Flips gravity if the player will touch a slope with enough momentum. Reverses forward and backward input if the player is flipped around by the slope.
		if MobjWillTouchFloor(mo)
			local speed = FixedHypot(GetSlopeLandingMomentum(mo, GetFloorSlope(mo)))
			local threshold = FixedMul(CEILING_SPEED_THRESHOLD, mo.scale)
			local slope = GetFloorSlope(mo)
			if not ((slope==nil) or (slope.flags & SL_NOPHYSICS) or (slope.zangle == 0))
				if abs(AngleFixed(GetFloorSlopeXY(mo, slope)) - AngleFixed(player.drawangle)) > 45*FRACUNIT
					player.flipdir = true
				end
				if speed >= threshold
					FlipGravity(mo)
					player.ceilingrun = true
					if WallRun
						//clears wallrun specific variables if present
						player.wallruntype = 0
						player.stoptimer = false
						player.fakewallrun = 0
						player.wallrun = nil
					end
					
					if (mo.state == S_PLAY_ROLL or mo.state == S_PLAY_WALLROLL) and player.flipdir and flip
						player.wasroll = 10
					end
					
					if not (twodlevel or mo.flags2 & MF2_TWOD)
						Set3DMoveVars(player)
					end
				end
				//first ceilingrun tic input flipping, prevents falling off immediately due to inputting the incorrect direction
				if player.flipdir and player.ceilingrun
					if twodlevel or mo.flags2 & MF2_TWOD
						player.cmd.sidemove = -$
					elseif not flip
						player.cmd.forwardmove = -$
					else
						//actually flips player angle if not in 2d and using flipcam (2d flipcam sucks btw)
						if (mo.state ~= S_PLAY_ROLL) and (mo.state ~= S_PLAY_WALLROLL)
							if player.cmd.forwardmove < 0
								mo.angle = GetFloorSlopeXY(mo, slope)
							else
								mo.angle = GetFloorSlopeXY(mo, slope) + ANGLE_180
							end
						end
					end
				end
			end
		elseif player.ceilingrun
			player.falltimer = $+1
			if (player.pflags & PF_JUMPED) or mo.state == S_PLAY_ROLL
				player.falltimer = 2
				if mo.state == S_PLAY_ROLL
					if WallRun
						//if wallrun is enabled, sets the same variables that would be applied from a halfpipe (requires a little bit of a slope to work properly)
						if abs(player.lastslopeang) > ANG20
							player.fakewallrun = 1
							player.wallruntype = 2
							mo.state = S_PLAY_WALLROLL
							player.pflags = $ | PF_SPINNING
						end
					end
					if mo.eflags & MFE_VERTICALFLIP and not (twodlevel or mo.flags2 & MF2_TWOD)
						mo.momz = $ * 9/4//z-momentum boost when leaving a wall from rolling to compensate for weird slope transfer when upside down
					else
						mo.momz = $ * 7/6 //slope launch physics work a lot better when leaving the floor, if your ceilingrun state
						//is floor running (reverse sector/global gravity), multiply by less to avoid weird momentum glitches
					end
				end
			end
			
			//tells wallrun to reverse wall input direction if running off a halfpipe while using flipcam in ceilingrun 
			//sets the correct angle based on movement direction
			if WallRun and flip and not(twodlevel or mo.flags2 & MF2_TWOD)
				if player.powers[pw_justlaunched] == 2 and not player.wallflip
					mo.angle = player.lastslope
					player.wallflip = true
					if player.cmd.forwardmove >= 0
						mo.angle = $ + ANGLE_180
					end
				end
			end
			
			if player.falltimer == 2
				FlipGravity(mo)
				player.ceilingrun = false
			end
		end
		
		if not player.ceilingrun
			player.wasroll = 0
			player.flipdir = false
			player.falltimer = 0
		end
	end
end)

//applies modified movement physics while in ceilingrun state
addHook("PlayerThink", function(player)
	if HPU.active == false
		return
	end
	
	local mo = player.mo
	if (player.nmomx ~= nil) and (player.ceilingrun == true) and not (twodlevel or mo.flags2 & MF2_TWOD)
		mo.momx = player.nmomx
		mo.momy = player.nmomy
		player.rmomx = player.nrmomx
		player.rmomy = player.nrmomy
		player.cmomx = player.ncmomx
		player.cmomy = player.ncmomy
		player.speed = player.nspeed
	end
	//sets nmoms to the actual player speed for use in 2D mode's reversebutterslope
	player.nmomx = player.momx
	player.nmomy = player.momy
end)

addHook("ThinkFrame", function()
	if HPU.active == false
		return
	end
	
	for player in players.iterate do
		if player.ceilingrun == true
			//changes slope physics if in ceilingrun state so upside down downward slopes have an upwards slope effect and vice versa
			P_ReverseButteredSlope(player.mo, player.nmomx, player.nmomy)
		end
		//clears all "new" movement variables 
		player.nmomx = nil
		player.nmomy = nil
		player.nrmomx = nil
		player.nrmomy = nil
		player.ncmomx = nil
		player.ncmomy = nil
		player.nspeed = nil
	end
end)

/*
Wall Run
version 1.2.2
by Kaldrum
This script adds wall running functionality when a player leaves a halfpipe, giving them either a wallrunning or wallrolling state.
When in either state, the player can walljump to gain a burst of horizontal momentum away from the wall, and can use their ability following this jump.
In the running state, the player is able to freely run on the wall, with inputs down increasing downwards momentum and inputs upwards partially counteracting gravity, allowing for extended time on the wall.
The player is able to transition into their wallrolling state at any time as long as they are currently touching the wall.
*/

if WallRun
	return
end

rawset(_G, "WallRun", true)

freeslot(
"S_PLAY_WALLRUN",
"S_PLAY_WALLROLL",
"S_TAILSOVERLAY_WALLRUN",
"SPR2_WLRN",
"SPR2_TALW"
)

local DASHMODE_THRESHOLD = TICRATE * 3
local DASHMODE_MAX = TICRATE * 3 + 3
local multiwallrun = true

//Needs different animations than default ones.
states[S_PLAY_WALLRUN] = {
	sprite = SPR_PLAY,
	frame = SPR2_WLRN,
	tics = 2,
	nextstate = S_PLAY_WALLRUN
}

states[S_PLAY_WALLROLL] = {
	sprite = SPR_PLAY,
	frame = SPR2_ROLL,
	tics = 1,
	nextstate = S_PLAY_WALLROLL
}

states[S_TAILSOVERLAY_WALLRUN] = {
	sprite = SPR_PLAY,
	frame = SPR2_TALW|FF_SPR2MIDSTART,
	tics = 2,
	nextstate = S_TAILSOVERLAY_WALLRUN
}

spr2defaults[SPR2_WLRN] = SPR2_RUN
spr2defaults[SPR2_TALW] = SPR2_TAL6

//Obtains the object's angle required to face directly towards a wall. Accounts for both front and back sidedefs.
local function GetWallAngle(mo, line)
	local mobjAngle = R_PointToAngle2(line.v1.x, line.v1.y , line.v2.x, line.v2.y)
	if(#line.frontsector == #mo.subsector.sector)
		mobjAngle = $ + ANGLE_90
		mo.lineside = 0
	elseif (line.backsector ~= nil) and (#line.backsector == #mo.subsector.sector)
		mobjAngle = $ - ANGLE_90
		mo.lineside = 1
	else //just in case
		mobjAngle = $ + ANGLE_90
		mo.lineside = 0
	end
	return mobjAngle
end

//Gets the XY angle of the launch slope in order to move directly up the slope.
local function GetFloorSlopeXY(mo)
	local slope = mo.standingslope
	local zangle = slope.zangle
	local xyangle = slope.xydirection
	if (slope.flags & SL_NOPHYSICS) or (zangle == 0)
		return nil
	end
	if zangle < 0 
		xyangle = $ + ANGLE_180
	end
	if mo.eflags & MFE_VERTICALFLIP
		xyangle = $ + ANGLE_180
	end
	return xyangle
end

/* If touching a wall following a halfpipe launch:
	Gets the player angle of the wall the player is touching.
	Checks if the "wall" is a thok barrier showing a skybox instead of wall texture. If so, the player cannot enter a wallrun state.
	Checks the difference between the previous wall angle and the current one. If this difference is 45 degrees or under, sets the wallangle to the new one. If not, ignores it.
	Also stores the side of this wall.
	This prevents the sides of sectors from rotating the player 90 degrees when they move between them. If there is no angle stored, stores the current wall angle.
	Sets the proper state depending on how the player launches from the slope.
	Resets the wallrun state loss tier, ends fakewallrun, and replenishes wallrun (the variable) to 1.
*/
addHook("MobjMoveBlocked", function(mo, object, line)
	if (not mo.player) or (HPU.active == false)
		return
	end
	if mo.player.wallruntype > 0	
		local lineangle = GetWallAngle(mo, line)
		local lastside = nil
		
		if mo.lineside == 0
			if line.backsector ~= nil
				if(mo.z > line.backsector.ceilingheight) and ((line.frontside.toptexture == 0) or (line.backsector.ceilingpic == "F_SKY1"))
					return
				elseif(mo.z < line.backsector.floorheight) and ((line.frontside.bottomtexture == 0) or (line.backsector.floorpic == "F_SKY1"))
					return
				elseif (line.flags & ML_IMPASSIBLE) and (mo.z>line.backsector.floorheight) and (mo.z<line.backsector.ceilingheight)
					return
				end
			end
			lastside = #line.frontside
		elseif mo.lineside == 1
			if(mo.z > line.frontsector.ceilingheight) and (line.backside.toptexture == 0)
				return
			elseif(mo.z < line.frontsector.floorheight) and ((line.backside.bottomtexture == 0) or (line.frontsector.floorpic == "F_SKY1"))
				return
			elseif (line.flags & ML_IMPASSIBLE) and (mo.z>line.frontsector.floorheight) and (mo.z<line.frontsector.ceilingheight)
				return
			end
			lastside = #line.backside
		end

		if mo.player.wallangle == nil
			mo.player.wallangle = lineangle
			mo.player.lastside = lastside
		elseif abs(mo.player.wallangle - lineangle) <= ANGLE_45
			mo.player.wallangle = lineangle
			mo.player.lastside = lastside
		end

		if (mo.player.wallruntype == 1) and (abs(mo.player.drawangle - lineangle) <= ANG60)
			if mo.state ~= S_PLAY_WALLRUN
				mo.state = S_PLAY_WALLRUN
			end
		elseif mo.player.wallruntype == 2
			if mo.state ~= S_PLAY_WALLROLL
				mo.state = S_PLAY_WALLROLL
			end
			mo.player.pflags = $ | PF_SPINNING
		end
		
		mo.player.stoptimer = false
		mo.player.pflags = $ & ~ PF_JUMPED
		mo.player.wallrun = 1
		mo.player.fakewallrun = 0
		mo.player.exitwallrun = 0	
	end
end)

addHook("PlayerThink", function(player)
	if HPU.active == false
		return
	end
	local mo = player.mo
	local flip = player.pflags & PF_FLIPCAM
	
	if mo.standingslope
		local dir = GetFloorSlopeXY(mo)
		if dir ~= nil
			player.lastslope = dir
		end
	end

	//Launch handling, if the player launches off a halfpipe while rolling, they enter wallroll. If running (or any other state), they enter wallrun.
	if player.powers[pw_justlaunched] == 2
		if not(twodlevel or mo.flags2 & MF2_TWOD)
			if player.cmd.forwardmove ~= 0
				player.wrinputdir = player.cmd.forwardmove/abs(player.cmd.forwardmove)
			else
				player.wrinputdir = 1
			end
		else
			if player.cmd.sidemove ~= 0
				player.wrinputdir = player.cmd.sidemove/abs(player.cmd.sidemove)
			else
				player.wrinputdir = 1
			end
		end
		
		if flip and player.wallflip
			player.wrinputdir = -$
			player.wallflip = false
		end
		
		//tells moveblocked which wallrun state to assign, as well as puts the player into a limited wallrun state until they are able to touch the wall
		if mo.state ~= S_PLAY_ROLL
			player.wallruntype = 1
			mo.momz = $ * 4/3
			mo.state = S_PLAY_WALLRUN
		elseif mo.state == S_PLAY_ROLL
			player.wallruntype = 2
			mo.state = S_PLAY_WALLROLL
			player.pflags = $ | PF_SPINNING
		end
		player.fakewallrun = 1
		P_Thrust(mo, player.lastslope, FRACUNIT)
		player.pflags = $ & ~ PF_JUMPED
		if not (player.cmd.buttons & BT_JUMP)
			player.wjready = true
		end
	end
	
	if P_IsObjectOnGround(mo)
		if player.wallruntype ~= -1
			mo.rollangle = 0
			player.wallruntype = 0
		end
		player.stoptimer = false
		if mo.state == S_PLAY_WALLROLL
			mo.state = S_PLAY_ROLL
			player.pflags = $ | PF_SPINNING
		end
		player.fakewallrun = 0
	end
	
	//Removes the stored wall angle and sets the reentry timer to its end if the player has lost their reentry ability.
	if not player.wallruntype and not(mo.state == S_PLAY_WALLRUN or mo.state == S_PLAY_WALLROLL)
		player.wallangle = nil
		player.wallrun = nil
		player.wrinputdir = nil
		player.prevtextoff = nil
		player.prevrowoff = nil
		player.lastside = nil
		player.exitwallrun = 16
		mo.lineside = nil
		player.wallruntype = -1
	end
	
	//conveyor movement
	if player.wallrun and player.wallangle and (player.lastside == player.lastlastside) and player.prevtextoff ~= nil
		player.wcmomx = FixedMul(sides[player.lastside].textureoffset - player.prevtextoff , cos(player.wallangle + ANGLE_90))
		player.wcmomy = FixedMul(sides[player.lastside].textureoffset - player.prevtextoff , sin(player.wallangle + ANGLE_90))
		player.wcmomz = sides[player.lastside].rowoffset - player.prevrowoff
	else
		player.wcmomx = 0
		player.wcmomy = 0
		player.wcmomz = 0
		if player.pwcmomx == nil
			player.pwcmomx = 0
			player.pwcmomy = 0
			player.pwcmomz = 0
		end
	end
	
	if player.lastside
		player.prevtextoff = sides[player.lastside].textureoffset
		player.prevrowoff = sides[player.lastside].rowoffset
		player.lastlastside = player.lastside
	end	
	
	mo.momx = $ - player.pwcmomx + player.wcmomx
	mo.momy = $ - player.pwcmomy + player.wcmomy
	mo.momz = $ - player.pwcmomz + player.wcmomz
	
	//previous conveyor values
	player.pwcmomx = player.wcmomx
	player.pwcmomy = player.wcmomy
	player.pwcmomz = player.wcmomz
	
	//Makes sure the player is always facing into the wall
	if player.wallrun ~= nil and player.wallangle ~= nil
		player.drawangle = player.wallangle
	end
	
	//Wallrun wall movement
	if mo.state == S_PLAY_WALLRUN
		if player.charflags & SF_DASHMODE
			if player.wasdashmode == true
				player.dashmode = max(player.dashmode, DASHMODE_MAX)
			end
			player.dashmode = $+3 //gives an extra boost to dashmode if already in it as well as preserving values below threshold
		end
		//Gets player inputs for the first frame before they are stored
		if player.wallangle == nil
			if not(twodlevel or mo.flags2 & MF2_TWOD)
				player.forwardinput = player.cmd.forwardmove * player.wrinputdir
			else
				player.forwardinput = player.cmd.sidemove * player.wrinputdir
			end
		else
			if not(twodlevel or mo.flags2 & MF2_TWOD)
				P_Thrust(mo, player.wallangle - ANGLE_90, FRACUNIT * player.turninput / 50)
				//rollangle modifications look pretty jank right now, hopefully if proper sprites and models are made it'll look less weird
				if player.forwardinput > 0
					mo.rollangle = FixedAngle(FRACUNIT / player.turninput * -90 / 50)
				elseif player.forwardinput < 0
					mo.rollangle = ANGLE_180 + FixedAngle(FRACUNIT * player.turninput * 90 / 50)
				end
			else
				if player.forwardinput > 0
					mo.rollangle = 0
				elseif player.forwardinput < 0
					mo.rollangle = ANGLE_180
					//reverses drawangle in 2D as well as rollangle. looks weird with models, not sure why there is an inconsistency
					player.drawangle = player.wallangle + ANGLE_180
				end
			end
		end
		P_SetObjectMomZ(mo, abs(P_GetMobjGravity(mo)) * player.forwardinput / 100, true)
		
		//Wall spin, should add more character exclusions at some point to increase compatibility with other mods
		if (player.cmd.buttons & BT_SPIN) and not (mo.skin == "fang" or mo.skin == "amy")
			mo.state = S_PLAY_WALLROLL
			player.wallruntype = 2
			player.pflags = $ | PF_SPINNING
			S_StartSound(mo, 17)
			mo.rollangle = 0
		end
	end
	
	//Walljump button handling
	if (player.cmd.buttons & BT_JUMP) and (player.wjready == true)
		player.wjready = false
		player.walljump = true
	elseif (player.cmd.buttons & BT_JUMP)
		player.wjready = false
		player.walljump = false
	elseif mo.state == S_PLAY_WALLRUN
		player.wjready = true
		player.walljump = false
	elseif mo.state == S_PLAY_WALLROLL
		player.wjready = true
		player.walljump = false
	else
		player.wjready = false
		player.walljump = false
	end

	//Walljump action, ends wallrun and ability to enter wallrun. Allows secondary jump actions.
	if player.walljump == true
		//if a walljump is attempted on the first frame of wallrun, must use last slope angle for wallangle purposes. 
		//Also reduces z-momentum, as it hasnt been effected by gravity yet.
		if player.wallangle == nil
			player.wallangle = player.lastslope
			mo.momz = $*3/4
		end
		P_InstaThrust(mo, player.wallangle + ANGLE_180, 20*FRACUNIT)
		mo.state = S_PLAY_JUMP
		S_StartSound(mo, 14)
		player.pflags = $ | PF_JUMPED
		player.pflags = $ & ~ PF_THOKKED
		mo.rollangle = 0
		//clears conveyor values, preventing the removal of their momentum next tic
		player.pwcmomx = 0
		player.pwcmomy = 0
		player.pwcmomz = 0
		//multiwallrun: if enabled, prevents the immediate wall run state loss from wall jumping. also rotates the player.
		if multiwallrun == true
			mo.angle = player.wallangle + ANGLE_180
			player.stoptimer = true
			if (twodlevel or mo.flags2 & MF2_TWOD)
				player.wrinputdir = -$
			end
			player.wallruntype = 1
		else		
			player.wallruntype = 0
		end
	end
	
	//Makes sure the state is cleared if still present when not on the wall.
	if (player.wallrun == 3) 
		if mo.state == S_PLAY_WALLRUN
			mo.state = S_PLAY_WALK
		elseif mo.state == S_PLAY_WALLROLL
			mo.state = S_PLAY_ROLL
			mo.momx = $ - FixedMul(abs($) , cos(player.wallangle))
			mo.momy = $ - FixedMul(abs($) , sin(player.wallangle))
		end
		mo.rollangle = 0
		mo.linexoff = nil
		mo.lineyoff = nil
		player.wallrun = nil
	elseif player.wallrun ~= nil
		if player.wallangle ~= nil
			if mo.state == S_PLAY_WALLROLL
				P_Thrust(mo, player.wallangle, 4*FRACUNIT)
			else
				P_Thrust(mo, player.wallangle, FRACUNIT)
			end
		end	
		if player.wallrun < 3
			player.wallrun = $ + 1
		end
	end	
	
	//small grace period to actually touch the wall when first off a halfpipe
	if player.fakewallrun
		player.fakewallrun = $ + 1
		if player.fakewallrun == 5
			player.fakewallrun = 0
			if mo.state == S_PLAY_WALLRUN
				mo.state = S_PLAY_WALK
			elseif mo.state == S_PLAY_WALLROLL
				mo.state = S_PLAY_ROLL
				mo.momx = $ - FixedMul(abs($) , cos(player.lastslope))
				mo.momy = $ - FixedMul(abs($) , sin(player.lastslope))
			end	
			player.wallruntype = 0
		else
			P_Thrust(mo, player.lastslope, FRACUNIT)
		end
	end
	
	//15 tic timer to remove ability to reenter wallrunning
	if player.exitwallrun < 15 and player.stoptimer == false and not player.wallrun
		player.exitwallrun = $ + 1
	end
	
	if player.exitwallrun == 15
		if player.cmd.forwardmove == 0 and not (twodlevel or mo.flags2 & MF2_TWOD)
			P_TryMove(mo, mo.x - (12 * cos(player.wallangle)), mo.y - (12* sin(player.wallangle)), mo.z)
		end
		player.wallruntype = 0
		player.exitwallrun = 16
	end
	
	//kinda lame implementation of this, seems like it works better now though with other fixes
	if player.charflags & SF_DASHMODE
		if player.dashmode > DASHMODE_THRESHOLD
			player.wasdashmode = true
		elseif player.dashmode < 84
			player.wasdashmode = false
		end
	end
end)

//Saves the players inputs for use in wall movement, but prevents them from acutally changing the player's direction.
addHook("PreThinkFrame", function()
	if HPU.active == false
		return
	end
	for player in players.iterate do	
		local mo = player.mo
		if mo.state == S_PLAY_WALLRUN
			if not(twodlevel or mo.flags2 & MF2_TWOD)
				player.forwardinput = player.cmd.forwardmove * player.wrinputdir
				player.turninput = player.cmd.sidemove * player.wrinputdir
				if (player.cmd.forwardmove ~= 0) and (player.forwardinput < 0) and (player.flipdir == false)
					player.cmd.forwardmove = 0
				end
				player.cmd.sidemove = 0
			else 
				//uses sidemove instead of forwardmove in 2d mode. disables sideways wall running
				player.forwardinput = player.cmd.sidemove * player.wrinputdir
				if (player.cmd.sidemove ~= 0) and (player.forwardinput < 0) and (player.flipdir == false)
					player.cmd.sidemove = 0
				end
				player.cmd.forwardmove = 0
			end
		elseif mo.state == S_PLAY_WALLROLL
			if not(twodlevel or mo.flags2 & MF2_TWOD)
				player.cmd.forwardmove = 0
			else
				player.cmd.sidemove = 0
				player.cmd.forwardmove = 0
			end	
			player.forwardinput = 0
			player.turninput = 0
		else
			player.forwardinput = 0
			player.turninput = 0
		end
	end
end)

addHook("MapChange", function(map)
	//add "Lua.multiwallrun = active" to a level header to enable subsequent wall runs following a walljump
	if mapheaderinfo[map].multiwallrun == "active"
		multiwallrun = true
	else
		multiwallrun = false
	end
end)

addHook("NetVars", function(net)
	multiwallrun = net(multiwallrun)
end)

	
/*TO DO:
	*add actual animations for wall running
	*probably more stuff idk
*/

/*
Animation 
version 1.0.1
by Sonic Team Junior
modified by Kaldrum
Pretty much direct source code ports to fix Tails' and Metal Sonic's overlays when in the wallrun state.
*/

if HPUAnim
	return
end

rawset(_G, "HPUAnim", true)

//source code but wallrun allows tails' tails to display. not needed if custom sprites are made
local function P_DoTailsOverlay(player, tails)
	local mo = player.mo
	local onground = P_IsObjectOnGround(mo)
	local horizangle = player.drawangle
	if (mo.flags2 & MF2_TWOD or twodlevel) and (mo.state == S_PLAY_WALLRUN) and (player.wallangle ~= nil) and player.forwardinput < 0
		horizangle = $ + ANGLE_180
	end
	local zoffs = 0
	local backwards = -1 * FRACUNIT
	local doswim = (player.panim == PA_ABILITY) and (player.mo.eflags & MFE_UNDERWATER)
	local doroll = (player.panim == PA_ROLL) or (player.panim == PA_JUMP and not (player.charflags & SF_NOJUMPSPIN)) or doswim or mo.state == S_PLAY_WALLROLL
	local rollangle = 0
	local panimchange = nil
	local ticnum = 0
	local chosenstate = 0
	
	if not tails.skin
		tails.skin = mo.skin
		tails.state = S_TAILSOVERLAY_STAND
		tails.movecount = -1
	end
	
	panimchange = (tails.movecount ~= player.panim)
	
	if doroll
		local testval = 0
		local zdist = 0
		if player.speed < FRACUNIT
			testval = FRACUNIT
		else
			testval = FixedMul(player.speed, cos(horizangle - R_PointToAngle2(0,0, player.rmomx, player.rmomy)))
			if testval < FRACUNIT
				testval = FRACUNIT
			end
		end
		if doswim
			zdist = mo.momz<<1
		elseif onground and not mo.reactiontime
			zdist = mo.z - tails.threshold
		else
			zdist = mo.momz
		end
		
		rollangle = R_PointToAngle2(0,0, testval, -P_MobjFlip(mo) * zdist)
		
		if not doswim
			zoffs = 3*FRACUNIT + 12 * sin(rollangle)
			backwards = -12 * cos(rollangle)
		end
	elseif player.panimn == PA_RUN or mo.state == S_PLAY_WALLRUN
		backwards = -5*FRACUNIT
	elseif player.panim == PA_SPRING or player.panim == PA_JUMP
		zoffs = $ + 4*FRACUNIT
		backwards = $ / 2
	elseif player.panim == PA_PAIN
		backwards = $ /16
	elseif mo.state == S_PLAY_GASP
		backwards = $ / 16
		zpffs = $ + 12*FRACUNIT
	elseif mo.state == S_PLAY_EDGE
		backwards = $ / 16
		zoffs = 3*FRACUNIT
	elseif player.panim == PA_ABILITY2
		zoffs = -7*FRACUNIT
		backwards = -9*FRACUNIT
	elseif player.panim == PA_ABILITY
		backwards = -5*FRACUNIT
	end
	
	if doroll
		local add = nil
		if rollangle > ANGLE_180
			add = 2
		else
			add = 0
		end
		
		if add
			rollangle = InvAngle(rollangle)
		end
		
		rollangle = $ + ANG15
		
		if rollangle > ANG60
			chosenstate = S_TAILSOVERLAY_PLUS60DEGREES + add
		elseif rollangle > ANG30
			chosenstate = S_TAILSOVERLAY_PLUS30DEGREES + add
		else
			chosenstate = S_TAILSOVERLAY_0DEGREES
		end
		
	elseif player.panim == PA_SPRING or player.panim == PA_JUMP
		chosenstate = S_TAILSOVERLAY_MINUS60DEGREES
	elseif player.panim == PA_FALL or player.panim == PA_RIDE
		chosenstate = S_TAILSOVERLAY_PLUS60DEGREES
	elseif player.panim == PA_PAIN
		chosenstate = S_TAILSOVERLAY_PAIN
	elseif mo.state == S_PLAY_GASP
		chosenstate = S_TAILSOVERLAY_GASP
	elseif mo.state == S_PLAY_EDGE
		chosenstate = S_TAILSOVERLAY_EDGE
	elseif player.panim == PA_DASH
		chosenstate = S_TAILSOVERLAY_DASH
	elseif player.panim == PA_RUN
		chosenstate = S_TAILSOVERLAY_RUN
	elseif mo.state == S_PLAY_WALLRUN
		chosenstate = S_TAILSOVERLAY_WALLRUN
	elseif player.panim == PA_WALK
		if not onground or mo.state == S_PLAY_SKID
			chosenstate = S_TAILSOVERLAY_PLUS30DEGREES
		elseif player.speed >= FixedMul(player.runspeed/2, mo.scale)
			chosenstate = S_TAILSOVERLAY_0DEGREES
		else
			chosenstate = S_TAILSOVERLAY_MINUS30DEGREES
		end
	elseif mo.sprite2 == SPR2_FLY
		chosenstate = S_TAILSOVERLAY_FLY
	elseif mo.sprite2 == SPR2_SWIM
		chosenstate = S_TAILSOVERLAY_FLY
	elseif mo.sprite2 == SPR2_TIRE
		chosenstate = S_TAILSOVERLAY_TIRE
	elseif player.panim == PA_ABILITY2
		chosenstate = S_TAILSOVERLAY_PLUS30DEGREES
	elseif player.panim == PA_IDLE
		chosenstate = S_TAILSOVERLAY_STAND
	else
		chosenstate = S_INVISIBLE
	end
	
	if panimchange
		tails.sprite2 = -1
		tails.state = chosenstate
	else
		if tails.state ~= chosenstate
			if states[chosenstate].sprite == SPR_PLAY
				tails.frame = states[chosenstate].frame & FF_FRAMEMASK
			end
			
			tails.state = chosenstate
		end
	end
	
	if player.fly1 ~= 0 and player.powers[pw_tailsfly] ~= 0 and not onground
		tails.state = chosenstate
	end
	
	if player.panim == PA_SPRING or player.panim == PA_FALL or mo.state == S_PLAY_RIDE
		if FixedDiv(abs(mo.momz), mo.scale) < 20<<16
			ticnum = 2
		else
			ticnum = 1
		end
	elseif player.panim == PA_PAIN
		ticnum = 2
	elseif mo.state == S_PLAY_GASP
		tails.tics = -1
	elseif mo.sprite2 == SPR2_TIRE
		if doswim
			ticnum = 2
		else
			ticnum = 4
		end
	elseif player.panim ~= PA_IDLE
		ticnum = mo.tics
	end
	
	if ticnum and (tails.tics > ticnum)
		tails.tics = ticnum
	end
	
	tails.color = mo.color
	tails.threshold = mo.z
	tails.movecount = player.panim
	tails.angle = horizangle
	P_SetScale(tails, mo.scale)
	tails.destscale = mo.destscale
	tails.radius = mo.radius
	tails.height = mo.height
	zoffs = FixedMul(zoffs, tails.scale)
	
	if mo.eflags & MFE_VERTICALFLIP
		tails.eflags = $ | MFE_VERTICALFLIP
		tails.flags2 = $ | MF2_OBJECTFLIP
		zoffs = player.mo.height - tails.height - zoffs
	else
		tails.eflags = $ & ~ MFE_VERTICALFLIP
		tails.flags2 = $ & ~ MF2_OBJECTFLIP
	end
	
	local x = mo.x + P_ReturnThrustX(tails, tails.angle, FixedMul(backwards, tails.scale))
	local y = mo.y + P_ReturnThrustY(tails, tails.angle, FixedMul(backwards, tails.scale))
	local z = mo.z + zoffs
	P_TeleportMove(tails, x, y, z)
end
			
//source code but wallrun also allows a jetfume to appear. slightly offset when rollangle makes the player face upside down
local function P_DoMetalJetFume(player, fume) 
	local FUME_SKINCOLORS =
	{
		SKINCOLOR_ICY,
		SKINCOLOR_SKY,
		SKINCOLOR_CYAN,
		SKINCOLOR_WAVE,
		SKINCOLOR_TEAL,
		SKINCOLOR_AQUA,
		SKINCOLOR_SEAFOAM,
		SKINCOLOR_MINT,
		SKINCOLOR_PERIDOT,
		SKINCOLOR_LIME,
		SKINCOLOR_YELLOW,
		SKINCOLOR_SANDY,
		SKINCOLOR_GOLD,
		SKINCOLOR_APRICOT,
		SKINCOLOR_SUNSET
	}
	local mo = player.mo
	local angle = player.drawangle
	local rollangle = mo.rollangle
	local dist = 0
	local heightoffset = 0
	if (mo.eflags & MFE_VERTICALFLIP)
		heightoffset = mo.height - (P_GetPlayerHeight(player)>>1)
	else
		heightoffset = (P_GetPlayerHeight(player)>>1)
	end
	
	local dashmode = min(player.dashmode, DASHMODE_MAX)
	local underwater = mo.eflags & MFE_UNDERWATER
	
	if (mo.state ~= S_PLAY_WALLRUN and player.panim ~= PA_WALK and player.panim ~= PA_RUN and player.panim ~= PA_DASH)
		if fume.state ~= fume.info.spawnstate
			fume.state = fume.info.spawnstate
		end
		return
	end
	
	if underwater
		fume.movedir = $ + FixedAngle(FixedDiv(2*player.speed, 3 * mo.scale))
		fume.movefactor = $ + player.speed
		
		if fume.movefactor > FixedDiv(2*player.normalspeed, 3 *mo.scale)
			local i = 0
			local radiusv = 4*FRACUNIT
			local radiusx = P_ReturnThrustX(mo, angle, -mo.radius)
			local radiusy = P_ReturnThrustY(mo, angle, -mo.radius)
			local factorx = P_ReturnThrustX(mo, angle + ANGLE_90, mo.scale)
			local factory = P_ReturnThrustY(mo, angle + ANGLE_90, mo.scale)
			local offseth = 0
			local offsetv = 0
			local x = 0
			local y = 0
			local z = 0
			for i = -1, 1, 2 
			do
				offseth = i*P_ReturnThrustX(fume, fume.movedir, radiusv)
				offsetv = i*P_ReturnThrustY(fume, fume.movedir, radiusv)
				x = mo.x + radiusx + FixedMul(offseth, factorx)
				y = mo.y + radiusy + FixedMul(offseth, factory)
				z = mo.z + heightoffset + offsetv
				P_SpawnMobj(x, y, z, MT_SMALLBUBBLE).scale = mo.scale>>1
			end
			fume.movefactor = 0
		end
		
		if player.panim == PA_WALK
			if fume.state ~= fume.info.spawnstate
				fume.threshold = 0
				fume.state = fume.info.spawnstate
			end
			return
		end
	end
	
	if(fume.state == fume.info.spawnstate)
		fume.state = fume.info.seestate
	end
	
	if (dashmode > DASHMODE_THRESHOLD) and (fume.state ~= fume.info.seestate)
		fume.destscale = mo.scale
		fume.flags2 = $ & ~ MF2_DONTDRAW
		fume.flags2 = $ | (mo.flags2 & MF2_DONTDRAW)
		
	else
		if dashmode == DASHMODE_THRESHOLD and dashmode > fume.movecount
			fume.state = fume.info.seestate
			P_SetScale(fume, mo.scale << 1)
		end
		
		fume.flags2 = (fume.flags2 & ~ MF2_DONTDRAW) | (mo.flags2 & MF2_DONTDRAW)
		fume.destscale = mo.scale + FixedDiv(player.speed, player.normalspeed)
		
		if underwater
			fume.destscale = $/6
		else
			fume.destscale = $/3
		end
		
		fume.color = FUME_SKINCOLORS[(dashmode * table.maxn(FUME_SKINCOLORS) / (DASHMODE_MAX + 1))+1]
		
		if underwater
			fume.frame = fume.frame & FF_FRAMEMASK | FF_ANIMATE | (P_RandomRange(0,9) * FF_TRANS10)
			fume.threshold = 1
		elseif fume.threshold
			fume.frame = (fume.frame & FF_FRAMEMASK) | fume.state.frame
			fume.threshold = 0
		end
	end
	fume.movecount = dashmode
	fume.flags2 = (fume.flags2 & ~ MF2_OBJECTFLIP) | (mo.flags2 & MF2_OBJECTFLIP)
	fume.eflags = (fume.eflags & ~ MFE_VERTICALFLIP) | (mo.eflags & MFE_VERTICALFLIP)
	
	dist = -mo.radius - FixedMul(fume.info.radius, fume.destscale - mo.scale/3)
	//gets offset from the player sprite when rolled
	fume.rollangle = rollangle
	local x = mo.x + P_ReturnThrustX(fume, angle, dist)
	local y = mo.y + P_ReturnThrustY(fume, angle, dist)
	local z = mo.z + heightoffset - (fume.height>>1)
	P_TeleportMove(fume, x, y, z)
	if player.normalspeed >=  player.normalspeed * 2
		P_SpawnGhostMobj(fume)
	end
end

addHook("FollowMobj", function(player, tails)
	if (not player.mo.skin == "henchman") or (active == false)
		return
	end
	P_DoTailsOverlay(player, tails)
	return true
end, MT_ENGIEOVERLAY)
				
addHook("FollowMobj", function(player, fume)
	if (not player.mo.skin == "metalsonic") or (active == false)
		return
	end
	P_DoMetalJetFume(player, fume)
	return true
end, MT_METALJETFUME)